from typing import Union, Type, Any, Literal, Callable, List, Optional
from pydantic import validator, Field
from shapely.geometry import MultiPolygon, MultiPoint, MultiLineString, shape, Polygon, Point, LineString, LinearRing
from shapely.geometry.base import BaseMultipartGeometry
from dataclasses import dataclass
from warnings import warn

from metagen.base import BaseModelArbitrary

# TODO: Geometry for raster
# TODO: Solve buffer for the case of single line or two points

@dataclass
class ShapeCollection:
    collection: List[BaseMultipartGeometry]

    @property
    def type(self):
        return set([shape.type for shape in self.collection])

    @property
    def convex_hull(self) -> Polygon:
        """Return convex hull of the collection"""
        if self.type.issubset({'Polygon', 'MultiPolygon'}):
            return MultiPolygon(self.collection).convex_hull
        elif  self.type.issubset({'LineString', 'MultiLineString', 'LinearRing'}):
            return MultiLineString(self.collection).convex_hull
        elif self.type.issubset({'Point', 'MultiPoint'}):
            return MultiPoint(self.collection).convex_hull


def feature_collection(interface: dict) -> Polygon:
    """returns polygon of externa boundary from feature collection"""
    collection = ShapeCollection([shape(feature['geometry']) for feature in interface['features']])
    return collection.convex_hull


def geometry_collection(interface: dict) -> Polygon:
    collection = ShapeCollection([shape(geometry) for geometry in interface['geometries']])
    return collection.convex_hull


def polygon(interface: dict) -> Polygon:
    return shape(interface)


def multipolygon(interface: dict) -> Polygon:
    return shape(interface).convex_hull


def point(interface: dict) -> Polygon:
    return shape(interface).convex_hull


convert_func = {'FeatureCollection': feature_collection,
                'Polygon': polygon,
                'MultiPolygon': multipolygon,
                'GeometryCollection': geometry_collection}


class Geometry(BaseModelArbitrary):
    """
    Helper class to extract geometry from inpur geodata.
    input:
        data: any type that have __geo_interface__ like shaply, geopandas etc. Works for geometry types
        FeatureCollection, GeometryCollections, Multipolygons, Polygon, . It do not works for heterogenic collection of shapes

    """
    data: Any

    @property
    def geom_type(self):
        return self.data.__geo_interface__['type']

    @property
    def convert_func(self) -> Callable:
        return convert_func[self.geom_type]

    @validator('data', pre=True)
    def has_geo_interface(cls, value):
        if hasattr(value, '__geo_interface__'):
            return value
        else:
            raise AttributeError('Input data has not __geo_interface__')

    def bbox(self) -> dict:
        warn('Method "geometry" called insted of "box". Pathner Place do not support plain bbox')
        return self.geometry()

    def geometry(self) -> dict:
        return self.convert_func(self.data.__geo_interface__).__geo_interface__


import json
from abc import ABC
from pathlib import Path
from typing import Union, Any, Callable, List
import copy
from dataclasses import dataclass
from uuid import UUID


# helpe class to define and create pipileines
@dataclass
class Process(ABC):
    """Abstrac factory for processes"""

    @classmethod
    def __call__(self, *args, **kwargs):
        pass


class PipeMeta(type):
    def __new__(cls, clsname: str, superclasses: tuple, clsdict: dict):
        stac = [value for attr, value in clsdict.items() if (isinstance(value, (Process, Callable)))
                and (not attr.__contains__('__'))]
        clsdict.update(stac=stac)
        return super().__new__(cls, clsname, superclasses, clsdict)


class Pipe(metaclass=PipeMeta):

    def __call__(self, value) -> Any:
        current = value
        for method in self.stac:
            current = method(current)
        return current

# helper functions
def create_file(path: Path) -> None:
    path.mkdir(parents=True, exist_ok=True)


def str2Path(path: Union[Path, str]) -> Path:
    if not isinstance(path, Path):
        return Path(path)
    return path


def is_exist(obj: Any) -> Any:
    if not hasattr(obj, 'exists'):
        raise AttributeError(f'Object {obj} has not "exist" method ')
    if not obj.parent.exists():
        raise ValueError(f'{obj} does not exist')
    return obj


def prepare_data_for_leaf(obj: dict) -> dict:
    new = obj.copy()
    data = new.pop('data')
    new.update(data)
    return new


def load_json(path: Union[str, Path], encoding='utf8'):
    with open(path, 'r', encoding=encoding) as file:
        return json.load(file)


def make_hash(obj):
    """ Makes a hash from a dictionary, list, tuple or set to any level, that contains
  only other hashable types (including any lists, tuples, sets, and
  dictionaries). """

    if isinstance(obj, (set, tuple, list)):
        return tuple([make_hash(e) for e in obj])

    elif not isinstance(obj, dict):

        return hash(obj)

    new_o = copy.deepcopy(obj)
    for k, v in new_o.items():
        new_o[k] = make_hash(v)

    return hash(tuple(frozenset(sorted(new_o.items()))))


def validate_list_uuid(values: Union[List[str], List[UUID]])-> List[UUID]:
    if not values:
        return values
    new = []
    for value in values:
        if isinstance(value, str):
            new.append(UUID(value))
        elif isinstance(value, UUID):
            new.append(value)
        else:
            raise ValueError(f'Validation error: {value} should be UUID or UUID string representation')
    return new


class Singleton(object):
    _instance = None
    def __new__(class_, *args, **kwargs):
        if not isinstance(class_._instance, class_):
            class_._instance = object.__new__(class_, *args, **kwargs)
        return class_._instance


import click
from metagen.main import CONFIG

@click.group()
@click.version_option()
def main():
    """
    Ptr-Metagen Congiguration
    """
    pass


@main.group('Register', invoke_without_command=True)
@click.option('--show', '-s', 'show', is_flag=True,
              help='Show actual Register configuration')
def config(show):
    """Metagen configuration"""
    if show:
        click.echo(CONFIG)


if __name__ == "__main__":
    main()


from abc import ABC


class Preset(ABC):
    pass


class PresetBackgroundMaps(Preset):
    """
    Preset for MapBackgound
    """
    CartoDB_LightNoLabels = {
        "key": "cartoDB_LightNoLabels",
        "type": "wmts",
        "options": {
            "url": "https://{s}.basemaps.cartocdn.com/rastertiles/light_nolabels/{z}/{x}/{y}{r}.png"
        }}


class PresetMapSynchronization(Preset):
    AllTrue = {"roll": True,
               "tilt": True,
               "range": True,
               "center": True,
               "heading": True,
               "boxRange": True}


class PresetMapSettingView(Preset):
    Europe = {"center": {"lat": 45.77, "lon": 14.77}, "boxRange": 2103667}



from pytest import fixture
from shapely.geometry import Polygon, MultiPolygon, GeometryCollection
from geopandas import GeoDataFrame

from metagen.utils import Geometry
from metagen.utils.geometry import ShapeCollection


@fixture(autouse=True)
def polygon1():
    return Polygon([(0, 0), (1, 1), (1, 0)])


@fixture(autouse=True)
def polygon2():
    return Polygon([(0, 0), (-1, -1), (-1, 0)])


@fixture(autouse=True)
def polygon3():
    return Polygon([(0, 0), (-1, -1), (0, -1)])


@fixture(autouse=True)
def multipolygon(polygon1, polygon2, polygon3):
    return MultiPolygon([polygon1, polygon2, polygon3])


@fixture(autouse=True)
def geometry_collection(polygon1, polygon2, polygon3):
    return GeometryCollection([polygon1, polygon2, polygon3])


@fixture(autouse=True)
def geodataframe_poly(polygon1, polygon2, polygon3):
    return GeoDataFrame({'col1': ['name1', 'name2', 'name3'], 'geometry': [polygon1, polygon2, polygon3]}, crs="EPSG:4326")


# test Geometry
def test_geometry_has_geo_interface(polygon1):
    assert Geometry(data=polygon1)

    try:
        Geometry(data=1)
        assert False
    except AttributeError:
        assert True


def test_geometry_from_geopandas(geodataframe_poly, multipolygon):
    geom = Geometry(data=geodataframe_poly)
    assert geom.geometry() == multipolygon.convex_hull.__geo_interface__


def test_geometry_from_geometry_collection(geometry_collection, multipolygon):
    geom = Geometry(data=geometry_collection)
    assert geom.geometry() == multipolygon.convex_hull.__geo_interface__


def test_geometry_from_polygon(polygon1):
    geom = Geometry(data=polygon1)
    assert geom.geometry() == polygon1.convex_hull.__geo_interface__


def test_geometry_from_multipolygon(multipolygon):
    geom = Geometry(data=multipolygon)
    assert geom.geometry() == multipolygon.convex_hull.__geo_interface__

# test Colletion
def test_collection_poly(polygon1, polygon2, polygon3, multipolygon):
    collection = ShapeCollection([polygon1, polygon2, polygon3])
    assert collection.convex_hull == multipolygon.convex_hull



from typing import Union, Dict, Type
from uuid import UUID
from abc import ABC
from pydantic import BaseModel

from metagen.base import LeafABC, BaseModelWithDynamicKey, set_key_from_input
from metagen.helpers import make_hash


class Component(BaseModel, ABC):
    pass

    def __hash__(self) -> int:
        return make_hash(self.dict())


class Filter(BaseModelWithDynamicKey):
    __root__: Dict[str, UUID]

    @classmethod
    def set(cls, key_name: str, key: Union[str, UUID, Type[LeafABC]]):
        uuid = set_key_from_input(key)
        return cls(**{key_name: uuid})




import pytest
from click.testing import CliRunner
from metagen.config.cli import config

def test_register_show():
    runner = CliRunner()
    result = runner.invoke(config, input=True)
    assert result.exit_code == 0
    assert result.output == 'Hello World!\n'



from pathlib import Path
import pytest
from pydantic import AnyUrl

from metagen.main import CONFIG
from metagen.importer import Importer


def test_importer_init():
    try:
        importer = Importer(**CONFIG.importer_setting.dict())
    except Exception as e:
        assert False

    assert isinstance(importer.path, Path)
    assert isinstance(importer.instance_url, AnyUrl)


from pydantic import BaseModel, Field, AnyUrl
import yaml
from typing import Literal, Optional
from pathlib import Path

BASE_CONFIG_FILE = Path(__file__).parent / 'config.yaml'


class RegisterConfig(BaseModel):
    registerName: Literal['pandas', 'dict']


class ImporterConfig(BaseModel):
    path: Optional[Path]
    instance_url: Optional[AnyUrl]
    host: Optional[AnyUrl]


class Config(BaseModel):
    register_setting: RegisterConfig = Field(default_factory=RegisterConfig)
    importer_setting: ImporterConfig = Field(default_factory=ImporterConfig)


def load_yaml(path: str) -> dict:
    with open(path, 'r') as file:
        return yaml.load(file, Loader=yaml.FullLoader)


def dump_yaml(path: Path, data: dict) -> None:
    with open(path, 'w') as file:
        file.write(yaml.dump(data, Dumper=yaml.Dumper))





import pytest
from pathlib import Path

from metagen.pipes import PathCheck


def test_path_check():
    cwd = str(Path.cwd())
    pathchek = PathCheck()
    path = pathchek(cwd)
    assert isinstance(path, Path)


from abc import ABC, abstractmethod
from uuid import UUID
from typing import Type, Any, List, Dict
from pydantic import BaseModel, Field

from metagen.base import LeafABC
from metagen.helpers import Singleton


class RegisterABC(BaseModel, ABC):

    @abstractmethod
    def get_elements(self) -> List[Type[LeafABC]]:
        pass

    @abstractmethod
    def add(self, element: Type[LeafABC]) -> None:
        pass

    @abstractmethod
    def update(self, attrName: str, value: Any)-> None:
        pass

    @abstractmethod
    def check_register(self, element: Type[LeafABC]) -> bool:
        pass

    @abstractmethod
    def get_by_uuid(self, uuid: UUID) -> Type[LeafABC]:
        pass

    class Config:
        arbitrary_types_allowed = True


# register
class DictRegister(RegisterABC, Singleton):
    uuid: dict = Field(default_factory=dict)

    @property
    def hashes(self):
        return {hash(element): element for element in self.uuid.values()}

    def get_elements(self) -> List[Type[LeafABC]]:
        return [element for element in self.uuid.values()]

    def add(self, element: Type[LeafABC]) -> None:
        if not self.check_register(element):
            self.uuid.update({str(element.key): element})
        else:
            raise ValueError(f'PTR element "{element.__class__.__name__}" with '
                             f'key: {element.key} and hash: {hash(element)} already exist')

    def update(self, attrName: str, value: Any) -> None:
        raise NotImplementedError()

    def check_register(self, element: Type[LeafABC]) -> bool:
        return hash(element) in self.hashes

    def get_by_uuid(self, uuid: str) -> Type[LeafABC]:
        if UUID(uuid):
            return self.uuid.get(uuid)

    def get_by_hash(self, hash: int) -> Type[LeafABC]:
        return self.hashes.get(hash)


class RegisterFactory(BaseModel):
    registers: Dict[str, Type[RegisterABC]] = Field(default_factory=dict)

    def add(self, registerName: str, registerType: Type[RegisterABC]) -> None:
        self.registers.update({registerName: registerType})

    def get(self, registerName: str, **ignore) -> Type[RegisterABC]:
        return self.registers[registerName]


register_factory = RegisterFactory()
register_factory.add(registerName='dict', registerType=DictRegister)



import pathlib
from setuptools import setup
from metagen import __version__

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="ptrmetagen",
    version=__version__,
    description="Package for generation of metastructures for Panter project",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/gisat-panther/ptr-metagen",
    author="Michal Opetal",
    author_email="michal.opletal@gisat.cz",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9"],
    packages=["metagen", "metagen.components", "metagen.utils", "metagen.config"],
    package_data = {'': ['metagen/config/config.yaml']},
    include_package_data=True,
    entry_points='''
       [console_scripts]
       metagen=metagen.config.cli:main
   ''',
    )



from metagen.helpers import validate_list_uuid
from uuid import UUID, uuid4


def test_uuid_list_validation():
    assert all(map(lambda x: isinstance(x, UUID), validate_list_uuid([uuid4(), uuid4()])))
    assert all(map(lambda x: isinstance(x, UUID), validate_list_uuid([str(uuid4()), str(uuid4())])))
